Require Import PCOFE.
Require Import PCM.

Generalizable Variables T U V.

Section Definitions.
  Context `{cT : cofe T}.

  Class topBI  := top  : T.
  Class botBI  := bot  : T.
  Class andBI  := and  : T -n> T -n> T.
  Class orBI   := or   : T -n> T -n> T.
  Class implBI := impl : T -n> T -n> T.
  Class scBI   := sc   : T -n> T -n> T.
  Class siBI   := si   : T -n> T -n> T.
  Class allBI  :=
    all  : forall `{pU : cofe U}, (U -n> T) -n> T.
  Class xistBI :=
    xist : forall `{pU : cofe U}, (U -n> T) -n> T.

  Section Lift.
    Context `{cU : cofe U} (f : T -n> T -n> T) (P Q : U -n> T).

    Definition lift_bin : U -n> T :=
      uncurry f <M< Mprod P Q.

  End Lift.
End Definitions.

  
Class ComplBI `{pcT : pcofe T} {BIT : topBI} {BIB : botBI}
      {BIA : andBI} {BIO : orBI} {BII : implBI} {BISC : scBI} {BISI : siBI}
      {BIAll : allBI} {BIXist : xistBI} :=
  mkCBI {
      top_true    :  forall P, P ⊑ top;
      bot_false   :  forall P, bot ⊑ P;
      and_self    :  forall P, P ⊑ and P P;
      and_projL   :  forall P Q, and P Q ⊑ P;
      and_projR   :  forall P Q, and P Q ⊑ Q;
      and_pord    :> Proper (pord ==> pord ==> pord) (and : ne2);
      and_impl    :  forall P Q R, and P Q ⊑ R <-> P ⊑ impl Q R;
      impl_pord   :> Proper (pord --> pord ++> pord) (impl : ne2);
      or_injL     :  forall P Q, P ⊑ or P Q;
      or_injR     :  forall P Q, Q ⊑ or P Q;
      or_self     :  forall P, or P P ⊑ P;
      or_pord     :> Proper (pord ==> pord ==> pord) (or : ne2);
      sc_comm     :> Commutative (sc : ne2);
      sc_assoc    :> Associative (sc : ne2);
      sc_top_unit :  forall P, sc top P == P;
      sc_pord     :> Proper (pord ==> pord ==> pord) (sc : ne2);
      sc_si       :  forall P Q R, sc P Q ⊑ R <-> P ⊑ si Q R;
      si_pord     :> Proper (pord --> pord ++> pord) (si : ne2);
      all_R      U `{cU : cofe U} :
        forall P (Q : U -n> T), (forall u, P ⊑ Q u) <-> P ⊑ all Q;
      all_pord   U `{cU : cofe U} (P Q : U -n> T) :
        (forall u, P u ⊑ Q u) -> all P ⊑ all Q;
      xist_L     U `{cU : cofe U} :
        forall (P : U -n> T) Q, (forall u, P u ⊑ Q) <-> xist P ⊑ Q;
      xist_sc    U `{cU : cofe U} :
        forall (P : U -n> T) Q, sc (xist P) Q ⊑ xist (lift_bin sc P (umconst Q));
      xist_pord  U `{cU : cofe U} (P Q : U -n> T) :
        (forall u, P u ⊑ Q u) -> xist P ⊑ xist Q
    }.

Arguments topBI  : default implicits.
Arguments botBI  : default implicits.
Arguments andBI  T {_ _}.
Arguments orBI   T {_ _}.
Arguments implBI T {_ _}.
Arguments scBI   T {_ _}.
Arguments siBI   T {_ _}.
Arguments allBI  T {_ _}.
Arguments xistBI T {_ _}.
Arguments ComplBI T {_ _ _ _ _ _ _ _ _ _ _ _ _ _}.

Class Valid (T : Type) `{pcmT : pcofe T} :=
  mkValid
    { valid : T -> Prop;
      valid_top (t t' : T) (HV : valid t) : t' ⊑ t
    }.

Class Later (T : Type) `{pcT : pcofe T} {vT : Valid T} :=
  mkLater
    { later : T -m> T;
      later_mon (t : T) : t ⊑ later t;
      later_contr : contractive later;
      loeb (t : T) (HL : later t ⊑ t) : valid t
    }.

Arguments mkValid {_ _ _ _ _ _} _ _.
Arguments mkLater {_ _ _ _ _ _ _} _ _ _ _.

Notation " ▹ p " := (later p) (at level 20) : bi_scope.

Require Import UPred.

Section UPredLater.
  Context Res `{preoRes : preoType Res}.
  Local Obligation Tactic := intros; resp_set || eauto with typeclass_instances.

  Global Program Instance valid_up : Valid (UPred Res) :=
    mkValid (fun p : UPred Res => forall n r, p n r) _.
  Next Obligation.
    intros n r _; apply HV.
  Qed.

  Global Instance later_up_mon : Proper (pord ==> pord) later_up.
  Proof.
    intros p q Hpq [| n] r; [intros; exact I | simpl; apply Hpq].
  Qed.

  Global Program Instance later_upred : Later (UPred Res) :=
    mkLater m[(later_up)] _ _ _.
  Next Obligation.
    intros [| n] r Ht; [exact I | simpl].
    rewrite Le.le_n_Sn; assumption.
  Qed.
  Next Obligation.
    intros n p q Hpq [| m] r HLt; simpl; [tauto |].
    apply Hpq; auto with arith.
  Qed.
  Next Obligation.
    intros n r; induction n.
    - apply HL; exact I.
    - apply HL, IHn.
  Qed.

End UPredLater.

Section UPredBI.
  Context Res `{pcmRes : PCM Res}.
  Local Open Scope pcm_scope.
  Local Obligation Tactic := intros; eauto with typeclass_instances.

  (* Standard interpretations of propositional connectives. *)
  Global Program Instance top_up : topBI (UPred Res) := up_cr (const True).
  Global Program Instance bot_up : botBI (UPred Res) := up_cr (const False).

  Global Program Instance and_up : andBI (UPred Res) :=
    n[(fun P => n[(fun Q => mkUPred (fun n r => P n r /\ Q n r) _)])].
  Next Obligation.
    intros n m r1 r2 HLe HSub; rewrite HSub, HLe; tauto.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ n r; simpl; now rewrite EQQ.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ m r Hlt; simpl.
    split; intros; now (split; [| apply EQQ]).
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q n r; simpl; now rewrite EQP.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q m r Hlt; simpl.
    split; intros; now (split; [apply EQP |]).
  Qed.

  Global Program Instance or_up : orBI (UPred Res) :=
    n[(fun P => n[(fun Q => mkUPred (fun n r => P n r \/ Q n r) _)])].
  Next Obligation.
    intros n m r1 r2 HLe HSub; rewrite HSub, HLe; tauto.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ n r; simpl; now rewrite EQQ.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ m r Hlt; simpl.
    split; now (intros [HP | HQ]; [left | right; apply EQQ]).
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q n r; simpl; now rewrite EQP.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q m r Hlt; simpl.
    split; now (intros [HP | HQ]; [left; apply EQP | right]).
  Qed.

  Global Program Instance impl_up : implBI (UPred Res) :=
    n[(fun P => n[(fun Q =>
      mkUPred (fun n r => forall m r', m <= n -> r ⊑ r' -> P m r' -> Q m r') _)])].
  Next Obligation.
    intros n m r1 r2 HLe HSub HImp k r3 HLe' HSub' HP.
    apply HImp; try (etransitivity; eassumption); assumption.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ n r; simpl; now setoid_rewrite EQQ.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ m r Hlt; simpl.
    split; intros; apply EQQ, H; now eauto with arith.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q n r; simpl; now setoid_rewrite EQP.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q m r Hlt; simpl.
    split; intros; apply H, EQP; now eauto with arith.
  Qed.

  (* BI connectives. *)
  Global Program Instance sc_up : scBI (UPred Res) :=
    n[(fun P => n[(fun Q =>
      mkUPred (fun n r => exists r1 r2, Some r1 · Some r2 == Some r /\ P n r1 /\ Q n r2) _)])].
  Next Obligation.
    intros n m r1 r2 HLe [rd HEq] [r11 [r12 [HEq' [HP HQ]]]].
    rewrite <- HEq', assoc in HEq; setoid_rewrite HLe; revert HEq.
    destruct (Some rd · Some r11) as [r11' |] eqn: HEq''; intros HEq;
      [| erewrite pcm_op_zero in HEq by apply _; contradiction].
    repeat eexists; [eassumption | | assumption].
    eapply uni_pred, HP; [reflexivity |].
    exists rd; rewrite HEq''; reflexivity.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ n r; simpl; now setoid_rewrite EQQ.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ m r Hlt; simpl.
    split; intros [r1 [r2 [EQr [HP HQ]]]]; exists r1; exists r2;
    (split; [assumption | split; [| apply EQQ]; now auto with arith]).
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q n r; simpl; now setoid_rewrite EQP.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q m r Hlt; simpl.
    split; intros [r1 [r2 [EQr [HP HQ]]]]; exists r1; exists r2;
    (split; [assumption | split; [apply EQP |]; now auto with arith]).
  Qed.

  Global Program Instance si_up : siBI (UPred Res) :=
    n[(fun P => n[(fun Q =>
      mkUPred (fun n r => forall m r' rr, Some r · Some r' == Some rr -> m <= n -> P m r' -> Q m rr) _)])].
  Next Obligation.
    intros n m r1 r2 HLe [r12 HEq] HSI k r rr HEq' HSub HP.
    rewrite comm in HEq; rewrite <- HEq, <- assoc in HEq'.
    destruct (Some r12 · Some r) as [r' |] eqn: HEq'';
      [| erewrite comm, pcm_op_zero in HEq' by apply _; contradiction].
    eapply HSI; [eassumption | etransitivity; eassumption |].
    eapply uni_pred, HP; [| exists r12; rewrite HEq'']; reflexivity.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ n r; simpl; now setoid_rewrite EQQ.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ m r Hlt; simpl.
    split; intros; eapply EQQ, H; now eauto with arith.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q n r; simpl; now setoid_rewrite EQP.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q m r Hlt; simpl.
    split; intros; eapply H, EQP; now eauto with arith.
  Qed.

  (* Quantifiers. *)
  Global Program Instance all_up : allBI (UPred Res) :=
    fun T eqT mT cT =>
      n[(fun R => mkUPred (fun n r => forall t, R t n r) _)].
  Next Obligation.
    intros n m r1 r2 HLe HSub HR t; rewrite HLe, <- HSub; apply HR.
  Qed.
  Next Obligation.
    intros R1 R2 EQR n r; simpl; now setoid_rewrite EQR.
  Qed.
  Next Obligation.
    intros R1 R2 EQR m r HLt; simpl.
    split; intros; now apply EQR.
  Qed.
  
  Global Program Instance xist_up : xistBI (UPred Res) :=
    fun T eqT mT cT =>
      n[(fun R => mkUPred (fun n r => exists t, R t n r) _)].
  Next Obligation.
    intros n m r1 r2 HLe HSub [t HR]; exists t; rewrite HLe, <- HSub; apply HR.
  Qed.
  Next Obligation.
    intros R1 R2 EQR n r; simpl; now setoid_rewrite EQR.
  Qed.
  Next Obligation.
    intros R1 R2 EQR m r HLt; simpl.
    split; intros [t HR]; exists t; now apply EQR.
  Qed.

  (* All of the above preserve all the props it should. *)
  Global Instance and_up_ord : Proper (pord ==> pord ==> pord) (and_up : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ n r; simpl.
    rewrite EQP, EQQ; tauto.
  Qed.

  Global Instance or_up_ord : Proper (pord ==> pord ==> pord) (or_up : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ n r; simpl.
    rewrite EQP, EQQ; tauto.
  Qed.

  Global Instance impl_up_ord : Proper (pord --> pord ++> pord) (impl_up : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ n r HP m r'.
    rewrite <- EQP, <- EQQ; apply HP.
  Qed.

  Global Instance sc_up_ord : Proper (pord ==> pord ==> pord) (sc_up : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ n r HH; simpl.
    setoid_rewrite <- EQP; setoid_rewrite <- EQQ; apply HH.
  Qed.

  Global Instance si_up_ord : Proper (pord --> pord ++> pord) (si_up : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ n r HP m r' rr.
    rewrite <- EQP, <- EQQ; apply HP.
  Qed.

  Global Program Instance bi_up : ComplBI (UPred Res).
  Next Obligation.
    intros n r _; exact I.
  Qed.
  Next Obligation.
    intros n r HC; contradiction HC.
  Qed.
  Next Obligation.
    intros n r; simpl; tauto.
  Qed.
  Next Obligation.
    intros n r [HP HQ]; assumption.
  Qed.
  Next Obligation.
    intros n r [HP HQ]; assumption.
  Qed.
  Next Obligation.
    split; intros HH n r.
    - intros HP m r' HLe HSub HQ; apply HH; split; [rewrite HLe, <- HSub |]; assumption.
    - intros [HP HQ]; eapply HH; eassumption || reflexivity.
  Qed.
  Next Obligation.
    intros n r HP; left; assumption.
  Qed.
  Next Obligation.
    intros n r HQ; right; assumption.
  Qed.
  Next Obligation.
    intros n r; simpl; tauto.
  Qed.
  Next Obligation.
    intros P Q n r; simpl.
    setoid_rewrite (comm (Commutative := pcm_op_comm _)) at 1.
    firstorder.
  Qed.
  Next Obligation.
    intros P Q R n r; split.
    - intros [r1 [rr [EQr [HP [r2 [r3 [EQrr [HQ HR]]]]]]]].
      rewrite <- EQrr, assoc in EQr.
      destruct (Some r1 · Some r2) as [r12 |] eqn: EQr';
        [| now erewrite pcm_op_zero in EQr by apply _].
      exists r12 r3; split; [assumption | split; [| assumption] ].
      exists r1 r2; split; [rewrite EQr'; reflexivity | split; assumption].
    - intros [rr [r3 [EQr [[r1 [r2 [EQrr [HP HQ]]]] HR]]]].
      rewrite <- EQrr, <- assoc in EQr; clear EQrr.
      revert EQr; destruct (Some r2 · Some r3) as [r23 |] eqn: EQ23; intros EQr;
        [| now erewrite comm, pcm_op_zero in EQr by apply _].
      exists r1 r23; split; [assumption | split; [assumption |] ].
      exists r2 r3; split; [rewrite EQ23; reflexivity | split; assumption].
  Qed.
  Next Obligation.
    intros n r; split.
    - intros [r1 [r2 [EQr [_ HP]]]].
      eapply uni_pred, HP; [| exists r1]; trivial.
    - intros HP; exists (pcm_unit _) r; split;
      [erewrite pcm_op_unit by apply _; reflexivity |].
      simpl; unfold const; tauto.
  Qed.
  Next Obligation.
    split; intros HH n r.
    - intros HP m r' rr EQrr HLe HQ; apply HH; rewrite <- HLe in HP.
      eexists; eexists; split; [eassumption | tauto].
    - intros [r1 [r2 [EQr [HP HQ]]]]; eapply HH; eassumption || reflexivity.
  Qed.
  Next Obligation.
    split.
    - intros HH n r HP u; apply HH; assumption.
    - intros HH u n r HP; apply HH; assumption.
  Qed.
  Next Obligation.
    intros n r HA u; apply H, HA.
  Qed.
  Next Obligation.
    split.
    - intros HH n r [u HP]; eapply HH; eassumption.
    - intros HH u n r HP; apply HH; exists u; assumption.
  Qed.
  Next Obligation.
    intros n t [t1 [t2 [EQt [[u HP] HQ]]]]; exists u t1 t2; tauto.
  Qed.
  Next Obligation.
    intros n r [u HA]; exists u; apply H, HA.
  Qed.

End UPredBI.

(* This class describes a type that can close over "future Us",
   thus making a nonexpansive map monotone *)
Class MonotoneClosure T `{pcT : pcofe T} :=
  { mclose : forall {U} `{pcU : pcofe U} {eU : extensible U},
               (U -n> T) -n> U -m> T;
    mclose_cl : forall {U} `{pcU : pcofe U} {eU : extensible U} (f : U -n> T) u,
                  mclose f u ⊑ f u;
    mclose_fw : forall {U} `{pcU : pcofe U} {eU : extensible U} (f : U -n> T) u t
                       (HFW : forall u' (HS : u ⊑ u'), t ⊑ f u'),
                  t ⊑ mclose f u
  }.
Arguments Build_MonotoneClosure {_ _ _ _ _ _} _ {_ _}.

Section MonotoneExt.
  Context B `{BBI : ComplBI B} {MCB : MonotoneClosure B}
          T `{pcT' : pcofe T} {eT : extensible T}.
  Local Obligation Tactic := intros; resp_set || mono_resp || eauto with typeclass_instances.

  Global Instance top_mm : topBI (T -m> B) := pcconst top.
  Global Instance bot_mm : botBI (T -m> B) := pcconst bot.

  Global Program Instance and_mm : andBI (T -m> B) :=
    n[(fun P => n[(fun Q => m[(lift_bin and P Q)])])].
  Next Obligation.
    intros t1 t2 Subt; simpl morph.
    apply and_pord; now rewrite Subt.
  Qed.
  
  Global Program Instance or_mm  : orBI  (T -m> B) :=
    n[(fun P => n[(fun Q => m[(lift_bin or P Q)])])].
  Next Obligation.
    intros t1 t2 Subt; simpl morph.
    apply or_pord; now rewrite Subt.
  Qed.

  Global Program Instance impl_mm : implBI (T -m> B) :=
    n[(fun P => n[(fun Q => mclose (lift_bin impl P Q))])].
  Next Obligation.
    intros Q1 Q2 EQQ; apply morph_resp; intros t.
    now simpl morph; rewrite EQQ.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ; apply met_morph_nonexp; intros t.
    now simpl morph; rewrite EQQ.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q; simpl morph; apply morph_resp; intros t.
    now simpl morph; rewrite EQP.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q; simpl morph; apply met_morph_nonexp; intros t.
    now simpl morph; rewrite EQP.
  Qed.

  Global Program Instance sc_mm  : scBI  (T -m> B) :=
    n[(fun P => n[(fun Q => m[(lift_bin sc P Q)])])].
  Next Obligation.
    intros t1 t2 Subt; simpl morph.
    apply sc_pord; now rewrite Subt.
  Qed.
  
  Global Program Instance si_mm : siBI (T -m> B) :=
    n[(fun P => n[(fun Q => mclose (lift_bin si P Q))])].
  Next Obligation.
    intros Q1 Q2 EQQ; apply morph_resp; intros t.
    now simpl morph; rewrite EQQ.
  Qed.
  Next Obligation.
    intros Q1 Q2 EQQ; apply met_morph_nonexp; intros t.
    now simpl morph; rewrite EQQ.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q; simpl morph; apply morph_resp; intros t.
    now simpl morph; rewrite EQP.
  Qed.
  Next Obligation.
    intros P1 P2 EQP Q; simpl morph; apply met_morph_nonexp; intros t.
    now simpl morph; rewrite EQP.
  Qed.

  Global Program Instance all_mm : allBI (T -m> B) :=
    fun U eqU mU cU =>
      n[(fun R => m[(fun t => all n[(fun u => R u t)])])].
  Next Obligation.
    intros t1 t2 EQt; apply morph_resp; intros u.
    now simpl morph; rewrite EQt.
  Qed.
  Next Obligation.
    intros t1 t2 EQt; apply met_morph_nonexp; intros u.
    now simpl morph; rewrite EQt.
  Qed.
  Next Obligation.
    intros t1 t2 Subt; apply all_pord; intros u; simpl morph.
    now rewrite Subt.
  Qed.
  Next Obligation.
    intros R1 R2 EQR t; simpl morph; apply morph_resp; intros u; apply EQR.
  Qed.
  Next Obligation.
    intros R1 R2 EQR t; simpl morph; apply met_morph_nonexp; intros u; apply EQR.
  Qed.
  
  Global Program Instance xist_mm : xistBI (T -m> B) :=
    fun U eqU mU cU =>
      n[(fun R => m[(fun t => xist n[(fun u => R u t)])])].
  Next Obligation.
    intros t1 t2 EQt; apply morph_resp; intros u.
    now simpl morph; rewrite EQt.
  Qed.
  Next Obligation.
    intros t1 t2 EQt; apply met_morph_nonexp; intros u.
    now simpl morph; rewrite EQt.
  Qed.
  Next Obligation.
    intros t1 t2 Subt; apply xist_pord; intros u; simpl morph.
    now rewrite Subt.
  Qed.
  Next Obligation.
    intros R1 R2 EQR t; simpl morph; apply morph_resp; intros u; apply EQR.
  Qed.
  Next Obligation.
    intros R1 R2 EQR t; simpl morph; apply met_morph_nonexp; intros u; apply EQR.
  Qed.

  (* All of the above preserve all the props it should. *)
  Global Instance and_mm_ord : Proper (pord ==> pord ==> pord) (and_mm : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ t; simpl morph.
    apply and_pord; [apply EQP | apply EQQ].
  Qed.

  Global Instance or_mm_ord : Proper (pord ==> pord ==> pord) (or_mm : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ t; simpl morph.
    apply or_pord; [apply EQP | apply EQQ].
  Qed.

  Global Instance impl_mm_ord : Proper (pord --> pord ++> pord) (impl_mm : ne2).
  Proof.
    intros P1 P2 SubP Q1 Q2 SubQ t; unfold flip in SubP. unfold impl_mm.
    apply mclose_fw; intros t' Subt; rewrite Subt; clear t Subt; simpl morph.
    etransitivity; [apply mclose_cl |].
    simpl morph; apply impl_pord; now rewrite ?SubP, <- ?SubQ.
  Qed.

  Global Instance sc_mm_ord : Proper (pord ==> pord ==> pord) (sc_mm : ne2).
  Proof.
    intros P1 P2 EQP Q1 Q2 EQQ t; simpl morph.
    apply sc_pord; [apply EQP | apply EQQ].
  Qed.

  Global Instance si_mm_ord : Proper (pord --> pord ++> pord) (si_mm : ne2).
  Proof.
    intros P1 P2 SubP Q1 Q2 SubQ t; unfold flip in SubP; unfold impl, impl_mm.
    apply mclose_fw; intros t' Subt; rewrite Subt; clear t Subt; simpl morph.
    etransitivity; [apply mclose_cl |].
    simpl morph; apply si_pord; now rewrite ?SubP, <- ?SubQ.
  Qed.

  Global Program Instance bi_mm : ComplBI (T -m> B).
  Next Obligation.
    intros t; apply top_true.
  Qed.
  Next Obligation.
    intros t; apply bot_false.
  Qed.
  Next Obligation.
    intros t; simpl morph; apply and_self.
  Qed.
  Next Obligation.
    intros t; simpl morph; apply and_projL.
  Qed.
  Next Obligation.
    intros t; simpl morph; apply and_projR.
  Qed.
  Next Obligation.
    split; intros HH t; simpl morph.
    - apply mclose_fw; intros t' Subt; specialize (HH t'); simpl morph in *.
      rewrite Subt, <- and_impl; assumption.
    - rewrite and_impl, (HH t); simpl morph; now rewrite mclose_cl.
  Qed.
  Next Obligation.
    intros t; simpl morph; apply or_injL.
  Qed.
  Next Obligation.
    intros t; simpl morph; apply or_injR.
  Qed.
  Next Obligation.
    intros t; simpl morph; apply or_self.
  Qed.
  Next Obligation.
    intros f1 f2 t; simpl morph; apply sc_comm.
  Qed.
  Next Obligation.
    intros f1 f2 f3 t; simpl morph; apply sc_assoc.
  Qed.
  Next Obligation.
    intros t; simpl morph; apply sc_top_unit.
  Qed.
  Next Obligation.
    split; intros HH t; simpl morph.
    - apply mclose_fw; intros t' Subt; specialize (HH t'); simpl morph in *.
      rewrite Subt, <- sc_si; assumption.
    - rewrite sc_si, (HH t); simpl morph; now rewrite mclose_cl.
  Qed.
  Next Obligation.
    split.
    - intros HH t; simpl morph; rewrite <- all_R; intros u; simpl morph; apply HH.
    - intros HH u t; specialize (HH t); simpl morph in *; rewrite <- all_R in HH.
      simpl morph in HH; apply HH.
  Qed.
  Next Obligation.
    intros t; apply all_pord; intros u; simpl morph; apply H.
  Qed.
  Next Obligation.
    split.
    - intros HH t; simpl morph; rewrite <- xist_L; intros u; simpl morph; apply HH.
    - intros HH u t; specialize (HH t); simpl morph in *.
      rewrite <- xist_L in HH; simpl morph in HH; apply HH.
  Qed.
  Next Obligation.
    intros t; simpl morph.
    rewrite xist_sc; eapply xist_pord; intros u; simpl morph.
    reflexivity.
  Qed.
  Next Obligation.
    intros t; apply xist_pord; intros u; apply H.
  Qed.

End MonotoneExt.

Section MonotoneLater.
  Context T U `{LT : Later T} `{pcU : pcofe U}.
  Local Obligation Tactic := intros; resp_set || eauto with typeclass_instances.

  Global Program Instance valid_mon : Valid (U -m> T) :=
    mkValid (fun f : U -m> T => forall u, valid (f u)) _.
  Next Obligation.
    intros u; apply valid_top, HV.
  Qed.

  Global Program Instance later_mon_morph : Later (U -m> T) :=
    mkLater m[(fun f => m[(fun u => later (f u))])] _ _ _.
  Next Obligation.
    intros u; simpl morph; apply later_mon.
  Qed.
  Next Obligation.
    intros n f g Hfg u; simpl morph; apply later_contr, Hfg.
  Qed.
  Next Obligation.
    intros u; apply loeb, HL.
  Qed.

End MonotoneLater.

Section MComplUP.

  Context V `{pV : preoType V}.
  Local Obligation Tactic := intros; try resp_set.

  Section Def.
    Context T `{pcT : pcofe T} {eT : extensible T}.

    Program Definition mclose_up : (T -n> UPred V) -n> T -m> UPred V :=
      n[(fun f => m[(fun t => mkUPred (fun n v => forall t', t ⊑ t' -> f t' n v) _)])].
    Next Obligation.
      intros n m v1 v2 HLe HSubv HT t' HSubt.
      rewrite HLe, <- HSubv; apply HT, HSubt.
    Qed.
    Next Obligation.
      intros t1 t2 EQt n v; split; intros HT t' Subt; apply HT; [| symmetry in EQt];
      rewrite EQt; assumption.
    Qed.
    Next Obligation.
      intros t1 t2 EQt m v HLt; split; intros HT t' Subt;
      (destruct n as [| n]; [now inversion HLt |]).
      - symmetry in EQt.
        assert (HH : f t' = S n = f (extend t' t1)) by (eapply f, extend_dist; eassumption).
        apply HH; [assumption |].
        eapply HT, extend_sub; eassumption.
      - assert (HH : f t' = S n = f (extend t' t2)) by (eapply f, extend_dist; eassumption).
        apply HH; [assumption |].
        eapply HT, extend_sub; eassumption.
    Qed.
    Next Obligation.
      intros t1 t2 Subt n v HT t' Subt'; apply HT; etransitivity; eassumption.
    Qed.
    Next Obligation.
      intros f1 f2 EQf u n v; split; intros HH u' HSub; apply EQf, HH, HSub.
    Qed.
    Next Obligation.
      intros f1 f2 EQf u m v HLt; split; intros HH u' HSub; apply EQf, HH; assumption.
    Qed.

  End Def.

  Global Program Instance MCl_up : MonotoneClosure (UPred V) :=
    Build_MonotoneClosure mclose_up.
  Next Obligation.
    intros n v HH; apply HH; reflexivity.
  Qed.
  Next Obligation.
    intros n v HH u' HSub; apply HFW; assumption.
  Qed.

End MComplUP.

(* The above suffice for showing that the eqn used in Rose actually forms a Complete BI.
   The following would allow for further monotone morphisms to be added. *)

Section MComplMM.
  Context B `{BBI : ComplBI B} {MCB : MonotoneClosure B}
          V `{pcV : pcofe V} {eT : extensible V}.
  Local Obligation Tactic := intros; resp_set || mono_resp || eauto with typeclass_instances.

  Section Def.
    Context U `{pcU : pcofe U} {eU : extensible U}.

    Program Instance extensible_prod : extensible (U * V) :=
      mkExtend (fun uv1 uv2 => (extend (fst uv1) (fst uv2), extend (snd uv1) (snd uv2))).
    Next Obligation.
      destruct vd as [ud vd]; destruct ve as [ue ve].
      destruct HD as [HDu HDv]; destruct HS as [HSu HSv].
      split; eapply extend_dist; eassumption.
    Qed.
    Next Obligation.
      destruct vd as [ud vd]; destruct ve as [ue ve].
      destruct HD as [HDu HDv]; destruct HS as [HSu HSv].
      split; eapply extend_sub; eassumption.
    Qed.

    Program Definition mclose_mm : (U -n> V -m> B) -n> U -m> V -m> B :=
      n[(fun f => mcurry (mclose n[(fun uv => f (fst uv) (snd uv))]))].
    Next Obligation.
      intros [u1 v1] [u2 v2] [EQu EQv]; simpl morph.
      unfold fst, snd in *; rewrite EQu, EQv; reflexivity.
    Qed.
    Next Obligation.
      intros f1 f2 EQf u v; simpl morph.
      apply (morph_resp mclose); clear u v; intros [u v]; simpl morph.
      apply EQf.
    Qed.
    Next Obligation.
      intros f1 f2 EQf u v; simpl morph.
      apply mclose; clear u v; intros [u v]; simpl morph.
      apply EQf.
    Qed.

  End Def.

  Global Program Instance MCl_mm : MonotoneClosure (V -m> B) := Build_MonotoneClosure mclose_mm.
  Next Obligation.
    intros v; simpl morph.
    rewrite mclose_cl; reflexivity.
  Qed.
  Next Obligation.
    intros v; simpl morph.
    apply mclose_fw; intros [u' v'] [HSubu HSubv]; simpl morph.
    rewrite HSubv; apply HFW; assumption.
  Qed.

End MComplMM.

Section Foo.
  Local Obligation Tactic := intros.

  Global Program Instance BI_valid T `{ComplBI T} : Valid T :=
    mkValid (fun t => top ⊑ t) _.
  Next Obligation.
    etransitivity; [apply top_true | assumption].
  Qed.

End Foo.
