(** Basic categorical definitions. The role of objects is played by
    the class [type] which is a type equipped with an equivalence
    relation.

    Then there are standard constructions defined using these objects
    (product, exponential, initial objects and functors at the end)
    and some of their properties are proved.
 *)

Require Export Coq.Program.Program.
Require Export Morphisms SetoidTactics.
(* Important to import last, to make sure equiv refers to SetoidClass.equiv *)
Require Export SetoidClass.

Generalizable Variables T U V W.

Notation "'mkType' R" := (@Build_Setoid _ R _) (at level 10).

(** A morphism between two types is an actual function together with a
    proof that it preserves equality. *)
Record morphism S T `{eqS : Setoid S} `{eqT : Setoid T} :=
  mkMorph {
    morph :> S -> T;
    morph_resp : Proper (equiv ==> equiv) morph}.

Arguments mkMorph [S T eqS eqT] _ _.
Arguments morph_resp [S T eqS eqT] _ _ _ _.

Infix "-=>" := morphism (at level 45, right associativity).
Notation "'s[(' f ')]'" := (mkMorph f _).
Ltac resp_set :=
  intros t1 t2 HEqt; repeat (intros ?); simpl in *; try rewrite !HEqt; reflexivity.

(** This seems just to be a package putting together all the
    ingredients of [type], i.e. the carrier set, the relation and the
    property that the relation is an equivalence relation. *)
Record eqType :=
  {eqtyp  :> Type;
   eqtype :  Setoid eqtyp}.
Instance eqType_proj_type {ET : eqType} : Setoid ET := eqtype _.
Definition fromType T `{eT : Setoid T} : eqType := Build_eqType T _.

Section Morphisms.
  Context `{eT : Setoid T} `{eU : Setoid U} `{eV : Setoid V}.

  (** The type of equivalence-preserving maps is again a type with
      equivalence, defined pointwise. *)
  Global Program Instance morph_type : Setoid (T -=> U) :=
    mkType (fun f g => forall x, f x == g x).
  Next Obligation.
    clear; split.
    - intros f x; reflexivity.
    - intros f g HS x; symmetry; apply HS.
    - intros f g h Hfg Hgh x; etransitivity; [apply Hfg | apply Hgh].
  Qed.

  (** The application of [morphsm] to its argument preserves equality
      in both the function and the argument. *)
  Global Instance equiv_morph :
    Proper (equiv ==> equiv ==> equiv) (morph T U).
  Proof.
    intros f g HEq x y HEq'; etransitivity; [apply HEq | apply g, HEq'].
  Qed.

  (** Definition of composition of morphisms. Note the different
      arrows, i.e., -=> and ->. *)
  Program Definition mcomp (f: U -=> V) (g: T -=> U) : (T -=> V) :=
    s[(f ∘ g)].
  Next Obligation.
    intros x y HEq; apply f, g; assumption.
  Qed.

  Local Obligation Tactic := intros; resp_set || program_simpl.

  (** Two specific morphisms, identity and constant maps. *)
  Program Definition mid : T -=> T := s[(fun x => x)].

  Program Definition mconst (u : U) : T -=> U := s[(const u)].

End Morphisms.

Infix "<<" := mcomp (at level 35).
Arguments mid T {eT}.
Arguments compose {_ _ _} _ _ _ / .
Arguments const {_ _} _ _ / .

Section MorphConsts.
  Context `{eT : Setoid T} `{eU : Setoid U} `{eV : Setoid V} `{eW : Setoid W}.

  (** Composition maps equal morphism to equal morphisms. *)
  Global Instance equiv_mcomp :
    Proper (equiv (A := U -=> V) ==> equiv ==> equiv) mcomp.
  Proof.
    intros f f' HEq g g' HEq' x; simpl; rewrite HEq, HEq'; reflexivity.
  Qed.

  (** Composition of morphisms is associative. *)
  Lemma mcomp_assoc (f: V -=> W) (g: U -=> V) (h: T -=> U) :
    f << (g << h) == (f << g) << h.
  Proof. intros x; reflexivity. Qed.

  (** Identity is left- and right- identity for composition *)
  Lemma mcomp_idR (f : U -=> V) :
    f << mid _ == f.
  Proof. intros x; reflexivity. Qed.
  Lemma mcomp_idL (f : U -=> V) :
    mid _ << f == f.
  Proof. intros x; reflexivity. Qed.

  (** Lift an ordinary function to a function on [type]'s. *)
  Definition lift2s (f : T -> U -> V) p q : (T -=> U -=> V) :=
    mkMorph (fun x => mkMorph (f x) (p x)) q.

  Local Obligation Tactic := intros; resp_set || eauto with typeclass_instances.

  (** Pre- and postcomposition, as equality-preserving maps *)
  Program Definition precomp (f : T -=> U) : (U -=> V) -=> (T -=> V) :=
    s[(fun g => g << f)].

  Program Definition postcomp (f : T -=> U) : (V -=> T) -=> (V -=> U) :=
    s[(fun g => f << g)].

  Definition m2 := T -=> U -=> V.
  Definition m3 := T -=> U -=> V -=> W.

  Definition coerce_m2 (f : m2) t u := f t u.
  Definition coerce_m3 (f : m3) t u v := f t u v.

End MorphConsts.

Coercion coerce_m2 : m2 >-> Funclass.
Coercion coerce_m3 : m3 >-> Funclass.

(*Instance Equiv_PropP : Equiv Prop := iff.
Instance type_PropP : type Prop := iff_equivalence.*)

Section SetoidProducts.
  Context `{eU : Setoid U} `{eV : Setoid V}.

  Definition prod_equiv (p1 p2 : U * V) :=
    fst p1 == fst p2 /\ snd p1 == snd p2.

  (** The product of two types is another type, with equality defined pointwise. *)
  Global Program Instance prod_type : Setoid (U * V) :=
    mkType prod_equiv.
  Next Obligation.
    split.
    - intros [u v]; split; reflexivity.
    - intros [u1 v1] [u2 v2] [Hu Hv]; split; symmetry; assumption.
    - intros [u1 v1] [u2 v2] [u3 v3] [Hu12 Hv12] [Hu23 Hv23]; split; etransitivity; eassumption.
  Qed.

  Global Instance prod_proper : Proper (equiv ==> equiv ==> equiv) (@pair U V).
  Proof.
    intros u1 u2 Hu v1 v2 Hv; split; assumption.
  Qed.

  Global Instance mfst_proper : Proper (equiv ==> equiv) (@fst U V).
  Proof.
    intros [u1 v1] [u2 v2] [Hu Hv]; assumption.
  Qed.

  Global Instance msnd_proper : Proper (equiv ==> equiv) (@snd U V).
  Proof.
    intros [u1 v1] [u2 v2] [Hu Hv]; assumption.
  Qed.

  Local Obligation Tactic := intros; resp_set || program_simpl.

  (** The projections are in fact proper morphisms, i.e. they preserve equality. *)
  Program Definition mfst : (U * V) -=> U := s[(fst)].

  Program Definition msnd : (U * V) -=> V := s[(snd)].

  Context `{eT : Setoid T}.

  (** Tupling is also a proper morphism, i.e. respects equality. *)
  Program Definition mprod (f: T -=> U) (g: T -=> V) : T -=> (U * V) :=
    s[(fun t => (f t, g t))].
  Next Obligation.
    add_morphism_tactic; intros.
    change ((f x, g x) == (f y, g y)).
    rewrite H; reflexivity.
  Qed.

  Lemma mprod_unique (f: T -=> U) (g: T -=> V) (h: T -=> U * V) :
    mfst << h == f -> msnd << h == g -> h == mprod f g.
  Proof.
    intros HL HR x; split; simpl; [rewrite <- HL | rewrite <- HR]; reflexivity.
  Qed.

End SetoidProducts.

Arguments mprod_unique [U eU V eV T eT f g h] _ _ _.

Section IndexedProducts.
  Context {I : Type} {P : I -> eqType}.

  (** Equality on the indexed product. Essentially the same as for binary products, i.e. pointwise. *)
  Global Program Instance prodI_type : Setoid (forall i, P i) :=
    mkType (fun p1 p2 => forall i, p1 i == p2 i).
  Next Obligation.
    split.
    - intros X x; reflexivity.
    - intros X Y HS x; symmetry; apply HS.
    - intros X Y Z HPQ HQR x; etransitivity; [apply HPQ | apply HQR].
  Qed.

  Local Obligation Tactic := intros; resp_set || program_simpl.

  (** Projection functions. *)
  Program Definition mprojI (i : I) : (forall i, P i) -=> P i :=
    s[(fun X => X i)].

  Context `{eT : Setoid T}.

  (** Tupling into the indexed products. *)
  Program Definition mprodI (f : forall i, T -=> P i) : T -=> (forall i, P i) :=
    s[(fun t i => f i t)].

  Lemma mprod_projI (f : forall i, T -=> P i) i : mprojI i << mprodI f == f i.
  Proof. intros X; reflexivity. Qed.

  (** Product with the projections is an actual product. *)
  Lemma mprodI_unique (f : forall i, T -=> P i) (h : T -=> forall i, P i) :
    (forall i, mprojI i << h == f i) -> h == mprodI f.
  Proof.
    intros HEq x i; simpl; rewrite <- HEq; reflexivity.
  Qed.

End IndexedProducts.

Section Exponentials.
  Context `{eT : Setoid T} `{eU : Setoid U} `{eV : Setoid V} `{eW : Setoid W}.

  Local Obligation Tactic := intros; resp_set || program_simpl.

  (** Composition of morphism _as a morphims_. *)
  Program Definition comps : (U -=> V) -=> (T -=> U) -=> T -=> V :=
    lift2s (fun f g => f << g) _ _.

  Program Definition muncurry (f : T -=> U -=> V) : T * U -=> V :=
    s[(fun p => f (fst p) (snd p))].

  (** Currying map, i.e. the exponential transpose. *)
  Program Definition mcurry (f : T * U -=> V) : T -=> U -=> V :=
    lift2s (fun t u => f (t, u)) _ _.

  (** f × g, i.e. 〈 f ∘ π, g ∘ π' 〉 *)
  Definition mprod_map (f : T -=> U) (g : V -=> W) := mprod (f << mfst) (g << msnd).

  (** Evaluation map. *)
  Program Definition meval : (T -=> U) * T -=> U :=
    s[(fun p => fst p (snd p))].

End Exponentials.

Section Exp_props.
  Context `{eT : Setoid T} `{eU : Setoid U} `{eV : Setoid V} `{eW : Setoid W}.

  (** (Λ(f) × id) ; eval = f, where Λ(f) is the exponential transpose. *)
  Lemma mcurry_com (f : T * U -=> V) : f == meval << (mprod_map (mcurry f) (mid _)).
  Proof. intros [a b]; reflexivity. Qed.

  (** Exponential transposes are unique. *)
  Lemma mcurry_unique (f : T * U -=> V) h :
    f == meval << (mprod_map h (mid _)) -> h == mcurry f.
  Proof. intros HEq a b; simpl; rewrite HEq; reflexivity. Qed.

End Exp_props.

(** The [unit] type is the terminal object, i.e., there's a unique
    morphism from any [Setoid] to [unit] *)
Section Terminals.
  Context `{eT : Setoid T}.

  Global Program Instance unit_type : Setoid unit := mkType (const (const True)).
  Next Obligation.
    split.
    - intros []; exact I.
    - intros [] [] _; exact I.
    - intros [] [] [] _ _; exact I.
  Qed.

  Definition mone_term : T -=> unit := mconst tt.

  Lemma mone_term_unique (f g : T -=> unit) : f == g.
  Proof.
    intros x; exact I.
  Qed.

End Terminals.

(** The empty [type] is the initial element, i.e. there is unique a
    morphism from it to any other [type]. *)
Section Initials.
  Context `{eT : Setoid T}.

  Inductive empty : Set := .
  Global Program Instance empty_type : Setoid empty := mkType (const (const False)).
  Next Obligation.
    split; intros [].
  Qed.

  Program Definition mzero_init : empty -=> T := s[(fun x => match x with end)].
  Next Obligation. now intros []. Qed.

  Lemma mzero_init_unique (f g : empty -=> T) : f == g.
  Proof. now intros []. Qed.

End Initials.

(** Subsets. *)
Section Subsetoid.
  Context `{eT : Setoid T} {P : T -> Prop}.
  Local Obligation Tactic := intros; resp_set || program_simpl.

  (** [type] of elements that satisfy the predicate P, i.e. a
      subset. Equality is inherited from the carrier type. *)
  Global Program Instance subset_type : Setoid {t : T | P t} :=
    mkType (fun x y => ` x == ` y).
  Next Obligation.
    split.
    - intros [x Hx]; simpl; reflexivity.
    - intros [x Hx] [y Hy] HS; simpl in *; symmetry; assumption.
    - intros [x Hx] [y Hy] [z Hz] Hxy Hyz; simpl in *; etransitivity; eassumption.
  Qed.

  (* XXX: not sure if needed, since the next map goes directly
  Global Instance proj1sig_proper :
    Proper (equiv (A := {t : T | P t}) ==> equiv) (@proj1_sig _ _).
  Proof. resp_set.
    intros [x Hx] [y Hy] HEq; simpl in *; assumption. Qed.*)

  (** Inclusion from the subset to the superset is an
      equality-preserving morphism. *)
  Program Definition mincl : {t : T | P t} -=> T := s[(@proj1_sig _ _)].

  (** If we have a morphism from B to A whose image is in the subset
      determined by P, then this morphism restricts to the one into
      the subset determined by P. *)
  Context `{eU : Setoid U}.
  Program Definition minherit (f : U -=> T) (HB : forall u, P (f u)) :
    U -=> {t : T | P t} := s[(fun u => exist P (f u) (HB u))].

  (** Inclusion from subset determined by P to the superset is a monomorphism. *)
  Lemma mforget_mono (f g : U -=> {t : T | P t}) :
    mincl << f == mincl << g -> f == g.
  Proof.
    intros HEq x; simpl; apply HEq.
  Qed.

End Subsetoid.

(** Lifting of a type by adding a new distinct element.

    This is used for several constructions: lookups in finite maps
    return function types, for instance.
 *)
Section Option.
  Context `{eT : Setoid T}.

  Definition opt_eq (x y : option T) :=
    match x, y with
      | None, None => True
      | Some x, Some y => x == y
      | _, _ => False
    end.

  Global Program Instance option_type : Setoid (option T) :=
    mkType opt_eq.
  Next Obligation.
    split.
    - intros [a |]; simpl; reflexivity.
    - intros [a |] [b |] HS; simpl in *; now trivial || symmetry; assumption.
    - intros [a |] [b |] [c |] Hab Hbc; simpl in *; try (exact I || contradiction); [].
      etransitivity; eassumption.
  Qed.

End Option.

Section OptDefs.
  Context `{eT : Setoid T} `{eU : Setoid U}.
  Local Obligation Tactic := intros; resp_set || program_simpl.

  Global Instance Some_proper : Proper (equiv ==> equiv) (@Some T).
  Proof. resp_set. Qed.

  Program Definition msome : T -=> option T := s[(@Some T)].

  Definition optbind (f : T -> option U) (ov : option T) : option U :=
    match ov with
      | Some v => f v
      | None   => None
    end.

  Program Definition moptbind : (T -=> option U) -=> option T -=> option U :=
    lift2s (optbind : (T -=> option U) -> option T -> option U) _ _.
  Next Obligation.
    intros [v1 |] [v2 |] EQv; try (contradiction EQv || exact I); [].
    unfold optbind; apply x, EQv.
  Qed.
  Next Obligation.
    intros f1 f2 EQf [x |]; [simpl morph | exact I].
    apply EQf.
  Qed.

End OptDefs.

(** Discrete equality (i.e., Coq's standard equality) for any
    type. Note it's not declared as an instance by default: this is by
    design to avoid the search finding this equality when it's not the
    right equality to use. Users may declare instances at specific
    types that *should* be treated as discrete. *)
Program Definition discreteType T : Setoid T := mkType (@eq T).

